/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "llmonit.hh"
#include "lsmp_con.hh"
#include "ParseLine.hh"
#include <time.h>
#include <sstream>
#include <fstream>
#include <cstdio>

using namespace thread;
using namespace std;

//======================================  timing utility
int 
gds_sleep(double dt) {
    timespec tv;
    if (dt <= 0) return 0;
    long dtsec = (long) dt;
    tv.tv_sec  = dtsec;
    tv.tv_nsec = (dt - dtsec) * 1e9;
    return nanosleep(&tv, NULL);
}

//======================================  ll_collect constructor.
int
main(int argc, const char* argv[]) {
    llmonit mon(argc, argv);
    mon.run();
}

//======================================  ll_collect constructor.
ll_collect::ll_collect(const std::string& node, const std::string& name)
  : _part(name), _node(node), _cons(0)
{
    ostringstream chan;
    chan << "C0:LLD-" << node << "_" << _part << "_";
    _late.set_channel(chan.str() + "LATENCY");
    _recv.set_channel(chan.str() + "RECEIVED");
    _length.set_channel(chan.str() + "LENGTH");
}

//======================================  ll_collect destructor.
ll_collect::~ll_collect(void) {
}

//======================================  Connect to shm partition.
void 
ll_collect::connect(void) {
    _cons = new LSMP_CON(_part.c_str());
    if (_cons < 0) {
	string msg = "Unable to connect to partition: ";
	msg += _part;
	throw runtime_error(msg);
    }
}

//======================================  Disconnect partition.
void 
ll_collect::disconnect(void) {
    delete _cons;
    _cons = 0;
}

//======================================  Collect statistics.
void* 
ll_collect::thread_entry(void) {
    connect();
    if (! _cons) return (void*)1;
    while(run()) {
	if (!_cons->get_buffer()) continue;
	Time now  = Now();
	Time tBuf = Time(_cons->getEvtID());
	long fLen = _cons->getLength();
	_cons->free_buffer();
	Interval late = now - tBuf;
	_recv.add(tBuf, 1.0);
	_late.add(tBuf, late);
	_length.add(tBuf, fLen);
    }
    return 0;
}

//=======================================  tswin constructor
ll_collect::_tswin::_tswin(Interval delta, Interval max_length)
  : _dt(delta), _window(max_length)
{}

//=======================================  tswin constructor
void 
ll_collect::_tswin::add(const Time& t, float v) {
    semlock lock(_mutex, semlock::wrlock);
    _window.fixedAppend(t, _dt, &v);
}

//=======================================  Clear the time series.
void 
ll_collect::_tswin::clear(void) {
    semlock lock(_mutex, semlock::wrlock);
    _window.Clear();
}

//=======================================  Extract a specified time interval
TSeries 
ll_collect::_tswin::extract(const Time& t, Interval dt) const {
    semlock lock(_mutex, semlock::wrlock);
    return _window.extract(t, dt);
}

//=======================================  Clear the time series.
void 
ll_collect::_tswin::set_channel(const std::string& chan) {
    _channel = chan;
}

//=======================================  Construct a writer
ll_writer::ll_writer(Interval dt, const collect_list& colist)
    : _dt(dt), _delay(32.0), _collect(colist), _lastGPS(0) {
}

//=======================================  Destroy a writer
ll_writer::~ll_writer(void) {
    _writer.close();
}

//=======================================  Construct a writer.
void 
ll_writer::open(const std::string& name, int runId) {
    _writer.setName(name);
    _writer.setRunID(runId);
}

//=======================================  Construct a writer.
void 
ll_writer::addChannel(const Time& t, const ll_collect::_tswin& cl) {
    TSeries ts(cl.extract(t, _dt));
    if (ts.empty() || ts.getEndTime() < t) ts.Clear(t, cl.dT());
    ts.extend(t + _dt);
    _writer.addRawSeries(cl.channel().c_str(), ts);
}

//=======================================  Set output frame path template
void 
ll_writer::set_delay(Interval delay) {
    _delay = delay;
}

//=======================================  Set output frame path template
void 
ll_writer::set_frame_path(const std::string& path) {
    string::size_type inx = path.find_last_of('/');
    if (inx == string::npos) {
	_pattern.set_prefix(path);
    } else {
	_pattern.set_directory(path.substr(0,inx));
	_pattern.set_prefix(path.substr(inx));
    }
}

//======================================  Define writer thread action.
//   Notes:
//   _lastGPS is the end-time of the previously written frame.
//   tEnd, endGPS is the start time of the last frame that can be written now.
void* 
ll_writer::thread_entry(void) {
    while(run()) {
	//------------------------------  Find last available frame time.
	int endGPS = (Now() - (_delay + _dt)).getS() ;
	endGPS -= endGPS % int(_dt);
	Time tEnd(endGPS);
	cout << "Start writer at GPS " << Now() << " endGPS = " << endGPS 
	     << " end of last frame = " << _lastGPS.getS() << endl;

	//------------------------------  Don't write all frames from t=0
	if (! _lastGPS) {
	    int first = Now().getS() + int(_dt);
	    first -= first%int(_dt);
	    _lastGPS = Time(first);
	}

	//------------------------------  Write all unwritten frames.
	while (_lastGPS <= tEnd) {
	    write(_lastGPS);
	    _lastGPS += _dt;
	}

	//------------------------------  Wait for data for the next frame
	Interval tWait = (_lastGPS + _dt + _delay) - Now();
	if (double(tWait) <= 0.0) continue;
	gds_sleep(tWait);
    }
    return 0;
}

//======================================  collect data and write a frame
void 
ll_writer::write(const Time& t) {
    _writer.buildFrame(t, _dt);
    for (const_collect_iter i=_collect.begin(); i != _collect.end(); ++i) {
	addChannel(t, (*i)->ref_recv());
	addChannel(t, (*i)->ref_long());
	addChannel(t, (*i)->ref_late());
    }
    _pattern.set_dt(int(_dt));
    _pattern.make_dir(_pattern.dir_name(t));
    string file = _pattern.file_path(t);
    string temp = _pattern.temp_path(t);
    _writer.open(temp.c_str(), true);
    _writer.writeFrame();
    _writer.close();
    rename(temp.c_str(), file.c_str());
    _writer.erase();
}

//======================================  Construct a monitor
llmonit::llmonit(int argc, const char* argv[])
    : _name("LowLatencyStats"), _RunID(1234), _dt(128), _writer(0)
{
    _config = "/etc/llldd/llshm_config";
    bool syntax = false;
    for (int i=1; i<argc; ++i) {
	string argi = argv[i];
	if (argi == "-c") {
	    if (i+1 == argc) syntax = true;
	    else             _config = argv[++i];
	}
	else if (argi == "--help") {
	    syntax = true;
	}
	else if (argi == "-o") {
	    if (i+1 == argc) syntax = true;
	    else             _outdir = argv[++i];
	}
	else {
	    cerr << "Unrecognized argument: " << argi << endl;
	    syntax = true;
	    break;
	}
    }
    if (syntax) {
	cerr << "Command line syntax" << endl;
	cerr << "  llmonit [-c <config-file>] [-o <output-dir>]" << endl;
	return;
    }
    configure();
}

//======================================  Destroy the monitor
llmonit::~llmonit(void) {
    shut();
}

//======================================  Configure a monitor
void 
llmonit::configure(void) {
    //----------------------------------  Get the host node name
    const int max_nodename = 64;
    char node[max_nodename];
    gethostname(node, max_nodename);
    node[max_nodename-1] = 0;
    string fixnode = node;
    while (fixnode.find('-') != string::npos) {
	fixnode.replace(fixnode.find('-'),1,"_");
    }

    //----------------------------------  Read the configuration.
    ParseLine pl(_config.c_str());
    while (pl.getLine() >= 0) {
	int nArg = pl.getCount();
	if (nArg == 0) continue;
	_collectors.push_back(new ll_collect(fixnode, pl[0]));
    }

    //----------------------------------  Set up the writer thread.
    _writer = new ll_writer(_dt, _collectors);
    string prefix = "C-LLMONIT_";
    prefix += fixnode;
    if (_outdir.empty()) {
	_outdir ="/usr1/jzweizig/llmonit/";
	_outdir += prefix + "-%5r";
    }

    _writer->pattern().set_prefix(prefix);
    _writer->pattern().set_directory(_outdir);
    _writer->open(_name, _RunID);
}

//======================================  Run monitor
void 
llmonit::run(void) {
    for (collect_iter i=_collectors.begin(); i!=_collectors.end(); ++i) {
	(*i)->start_thread();
    }
    _writer->start_thread();
    while (!sleep(20)) {
	int nBusy = 0;
	for (const_collect_iter i=_collectors.begin(); 
	     i != _collectors.end(); ++i) {
	    if ((*i)->busy()) nBusy++;
	}
	if (!nBusy) break;
    }
}

//======================================  Shut down all the threads
void 
llmonit::shut(void) {
    _writer->stop_thread();
    for (collect_iter i=_collectors.begin(); i!=_collectors.end(); ++i) {
	(*i)->stop_thread();
	delete *i;
	*i = 0;
    }
}
